/*
** File: model_admxrc6tadv8.c  
** Project: ADB3 core driver
** Purpose: ADM-XRC-6T-ADV8 code, including function to bootstrap device
**          context.
**
** (C) Copyright Alpha Data 2011
**
** Notes for REVISION 2 board:
**
** 1. Organization of Flash device U37 (all offsets in bytes):
**
**  0x0000000 - 0x1FFFFFF  1st image (strap 00), failsafe ADB3 PCI-E to OCP
**                         Bridge bitstream
**  0x2000000 - 0x3FFFFFF  2nd image (strap 01), default ADB3 PCI-E to OCP
**                         Bridge bitstream
**  0x4000000 - 0x5FFFFFF  3rd image (strap 10), currently unused
**  0x6000000 - 0x7FFFFFF  4th image (strap 11), currently unused
**
** If a 64 MiB Flash device is fitted, the 3rd and 4th images are not
** available.
** If a 128 MiB Flash device is fitted, the 3rd and 4th images are available.
**
** 'strap' is a 2-bit binary value representing resistor straps on the board.
** The normal build of the ADM-XRC-6T-ADV8 uses a strap of 01 so that the 2nd
** image is the default bitstream loaded, and the 1st image is the fallback
** bitstream that is loaded if the FPGA cannot be configured (i.e. DONE does
** not go high) from the 2nd image.
*/

#include <df.h>

#include "adb3.h"
#include "adb3_common.h"
#include "admxrc6tadv8.h"
#include "admxrc6tx_common.h"
#include "avr_common.h"
#include "avr_flash.h"
#include "coredma.h"
#include "device.h"
#include "model_boot.h"
#include "flash.h"
#include "flash_cfi.h"
#include "i2c_common.h"

typedef struct _SensorInfo {
  const char* pDescription;
  uint32_t capabilities;
  struct {
    int16_t intPart;    /* 16-bit integer part representing range [-32768, 32767] */
    uint16_t fracPart;  /* 16-bit fractional part, representing range [0, 1) */
  } error;
  uint32_t datatype;
  uint32_t unit;
  int32_t exponent;
} SensorInfo;

/* TO DO - error values */
static SensorInfo g_avrVoltageInfo[] = {
  { "3.3V supply rail",            0, { 0,  1311 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC0 */
  { "12V supply rail",             0, { 0,  1966 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC1 */
  { "5V supply rail",              0, { 0,  2359 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC2 */
  { "1V supply rail",              0, { 0,  3277 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC3 */
  { "1.5V supply rail",            0, { 0,  4325 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC4 */
  { "1.8V supply rail",            0, { 0,  6554 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC5 */
  { "2.5V supply rail",            0, { 0, 15729 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC6 */
  { "VDD for 156.25 MHz osc.",     0, { 0,  3277 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* AVR ADC7 */
  /* Following sensor is NOT present in board rev. 2 */
  { "VDD for 150 MHz osc.",        0, { 3,     0 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }  /* AVR ADC8 */
};

/* TO DO - error values */
static SensorInfo g_avrTempInfo[] = {
  { "uC internal temperature",     0, { 3,     0 }, CoreSensorDataDouble, CoreSensorUnitC,    0 }, /* Microcontroller internal temp */
  { "FPGA int temp. diode",        0, { 3,     0 }, CoreSensorDataDouble, CoreSensorUnitC,    0 }, /* Temperature sensor diode in FPGA */
  { "Optical xcvr ext. temp.",     0, { 3,     0 }, CoreSensorDataDouble, CoreSensorUnitC,    0 }  /* Temperature sensor in TMP412 chip near optical transceivers */
};

/* TO DO - error values */
static SensorInfo g_avrMiscInfo[] = {
  { "Time since manufacture",      0, { 0, 16384 }, CoreSensorDataDouble, CoreSensorUnitS,    0 }, /* Time since manufactured */
  { "Event count",                 0, { 0,     0 }, CoreSensorDataUInt32, CoreSensorUnitNone, 0 }  /* Number of elapsed events */
};

/* These correspond to 'g_avrVoltageInfo' above, to convert the raw ADC values into Volt */
static struct {
  unsigned int fieldOffset; /* Offset of field in sensor data block */
  int32_t factor;           /* Fixed point 6.26, where 0x04000000 => 1, 0x02000000 => 0.5, 0xFE000000 => -0.5 etc. */
} g_avrVoltageScaling[] = {
  {  8U, 0x17AAB  /* factor = 0.00144450f */ },
  { 10U, 0x40000  /* factor = 0.00390625f */ },
  { 12U, 0x17AAB  /* factor = 0.00144450f */ },
  { 14U, 0x06000  /* factor = 0.00036621f */ },
  { 16U, 0x17AAB  /* factor = 0.00144450f */ },
  { 18U, 0x17AAB  /* factor = 0.00144450f */ },
  { 20U, 0x17AAB  /* factor = 0.00144450f */ },
  { 22U, 0x17AAB  /* factor = 0.00144450f */ },
  { 24U, 0x17AAB  /* factor = 0.00144450f */ }
};

/* These correspond to 'g_avrTempInfo' above, to convert the raw ADC values into deg. C */
static struct {
  unsigned int fieldOffset; /* Offset of field in sensor data block */
  unsigned int fieldMask;   /* Value for masking shifted raw value */
  boolean_t bFieldSigned;   /* TRUE => interpret raw ADC as signed (two's complement) */
  int32_t factor;           /* Fixed point 12.20, where 0x00100000 => 1, 0x00080000 => 0.5, 0xFFF80000 => -0.5 etc. */
  int32_t offset;           /* Fixed point 16.16, where 0x00010000 => 1, 0x00008000 => 0.5, 0xFFFF8000 => -0.5 etc. */
} g_avrTempScaling[] = {
  { 26U, 0xFFFFU, FALSE, 0x10000, 0xFEEED99A, /* factor = 1/16, offset = -273.15 */ },
  { 28U, 0xFFF0U,  TRUE,  0x1000,          0, /* factor = 1/256, offset = 0 */ },
  { 30U, 0xFFF0U,  TRUE,  0x1000,          0, /* factor = 1/256, offset = 0 */ }
};

static CoreVpdStatus
readVpdData(
  Adb3CoreDeviceContext* pDevCtx,
  uint16_t address,
  size_t n,
  void* pBuffer)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint8_t* p = (uint8_t*)pBuffer;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  address &= 0x1FFU;
  while (n) {
    *p++ = dfPciMemRead8(hModel, &pModelRegs->model3.avr.vpd[address]);
    address++;
    n--;
  }
  return CoreVpdSuccess;
}

static boolean_t
verifyChecksum(
	Adb3CoreDeviceContext* pDevCtx,
  uint16_t vpdLength)
{
  CoreVpdStatus status;
  uint8_t buff[16];
  uint8_t sum = 0, checksum;
  uint16_t offset = 0;
  uint16_t remaining = vpdLength - 1;
  uint16_t i, chunk;

  if (vpdLength < 5) {
  	dfDebugPrint(0, ("*** verifyChecksum: invalid VPD length %lu(0x%lx)\n",
      (unsigned long)vpdLength, (unsigned long)vpdLength));
    pDevCtx->info.vpd.bChecksumError = TRUE;
    return FALSE;
  }

  /* Get checksum from VPD memory; always last byte of structure */
  status = readVpdData(pDevCtx, remaining, sizeof(checksum), &checksum);
  if (CoreVpdSuccess != status) {
  	dfDebugPrint(0, ("*** verifyChecksum: failed to read checksum from VPD area\n"));
    pDevCtx->info.vpd.bChecksumError = TRUE;
    return FALSE;
  }

  /* Compute checksum over all but last byte of structure */
  while (remaining) {
    chunk = (remaining > sizeof(buff)) ? sizeof(buff) : remaining;
    status = readVpdData(pDevCtx, offset, chunk, buff);
    if (CoreVpdSuccess != status) {
  	  dfDebugPrint(0, ("*** verifyChecksum: failed to data chunk from VPD area for checksumming\n"));
      pDevCtx->info.vpd.bChecksumError = TRUE;
      return FALSE;
    }

    for (i = 0; i < chunk; i++) {
      sum = sum + buff[i];
    }
    offset = offset + chunk;
    remaining = remaining - chunk;
  }
  sum = (uint8_t)-sum;

  if (sum == checksum) {
    return TRUE;
  } else {
  	dfDebugPrint(0, ("*** verifyChecksum: checksum invalid, computed 0x%02lx found 0x%02lx\n",
      (unsigned long)sum, (unsigned long)checksum));
    pDevCtx->info.vpd.bChecksumError = TRUE;
    return FALSE;
  }
}

static boolean_t
readVpd(
  Adb3CoreDeviceContext* pDevCtx)
{
  CoreVpdStatus status;
  VpdAdmxrc6tadv8Rev0* pVpd = &pDevCtx->info.vpd.pData->admxrc6tadv8.rev0;
  VpdAdmxrc6tadv8Rev0 vpdRaw;
  uint16_t vpdVersion = 0xffffU;
  uint16_t vpdLength = 0U;
  int i;

  status = readVpdData(pDevCtx, 0, sizeof(vpdVersion), &vpdVersion);
  if (CoreVpdSuccess != status) {
  	dfDebugPrint(0, ("*** readVpd: failed to read VPD version field\n"));
    return FALSE;
  }
  vpdVersion = dfLe16ToCpu(vpdVersion);
  status = readVpdData(pDevCtx, 2, sizeof(vpdLength), &vpdLength);
  if (CoreVpdSuccess != status) {
  	dfDebugPrint(0, ("*** readVpd: failed to read VPD length field\n"));
    return FALSE;
  }
  vpdLength = dfLe16ToCpu(vpdLength);
  switch (vpdVersion) {
  case 0:
    /* Ok */
    break;

  case 0xffffU:
  	dfDebugPrint(0, ("*** readVpd: invalid VPD version %lu(0x%lx)\n",
      (unsigned long)vpdVersion, (unsigned long)vpdVersion));
    return FALSE;

  default:
  	dfDebugPrint(0, ("+++ readVpd: unrecognised VPD version %lu(0x%lx)\n",
      (unsigned long)vpdVersion, (unsigned long)vpdVersion));
    break;
  }

  verifyChecksum(pDevCtx, vpdLength);

  status = readVpdData(pDevCtx, 0, sizeof(vpdRaw), &vpdRaw);
  if (CoreVpdSuccess != status) {
  	dfDebugPrint(0, ("*** readVpd: failed to block read VPD\n"));
    return FALSE;
  }
  DF_ZERO_OBJECT(pVpd);
  pVpd->version = vpdVersion;
  pVpd->length = vpdLength;
  pVpd->serial = dfLe32ToCpu(vpdRaw.serial);
  pVpd->modifications = dfLe32ToCpu(vpdRaw.modifications);
  pVpd->idelayRef = dfLe32ToCpu(vpdRaw.idelayRef);
  pVpd->mgt1Ref = dfLe32ToCpu(vpdRaw.mgt1Ref);
  pVpd->mgt2Ref = dfLe32ToCpu(vpdRaw.mgt2Ref);
  for (i = 0; i < 4; i++) {
    pVpd->sdram[i].order = dfLe8ToCpu(vpdRaw.sdram[i].order);
    pVpd->sdram[i].speed = dfLe16ToCpu(vpdRaw.sdram[i].speed);
    pVpd->sdram[i].link = dfLe16ToCpu(vpdRaw.sdram[i].link);
  }
  pVpd->adv8NotFitted = dfLe32ToCpu(vpdRaw.adv8NotFitted);
  pVpd->fpga.device = dfLe16ToCpu(vpdRaw.fpga.device);
  dfAssert(sizeof(char) == sizeof(uint8_t));
  dfStrCpy((char*)pVpd->fpga.speedGrade, DF_ARRAY_LENGTH(pVpd->fpga.speedGrade), (const char*)vpdRaw.fpga.speedGrade);
  dfStrCpy((char*)pVpd->fpga.stepping, DF_ARRAY_LENGTH(pVpd->fpga.stepping), (const char*)vpdRaw.fpga.stepping);
  pVpd->fpga.tempGrade = dfLe8ToCpu(vpdRaw.fpga.tempGrade);
  pVpd->pcbRev = dfLe8ToCpu(vpdRaw.pcbRev);
  pVpd->checksum = dfLe8ToCpu(vpdRaw.checksum);

  pDevCtx->info.vpd.bValid = TRUE;

  return TRUE;
}

static CoreSensorStatus
readSensorAvrVoltage(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int avrVoltageIndex,
  CoreSensorValue* pReading)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint16_t val16;
  int32_t factor, value;
  unsigned int fieldOffset;

  dfAssert(DF_ARRAY_LENGTH(g_avrVoltageInfo) == DF_ARRAY_LENGTH(g_avrVoltageScaling));

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  if (avrVoltageIndex < pDevCtx->model.admxrc6tadv8.avr.numVoltageSensor) {
    fieldOffset = g_avrVoltageScaling[avrVoltageIndex].fieldOffset;
    val16  = (uint16_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 0) << 0;
    val16 |= (uint16_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 1) << 8;
    val16 &= 0xFFFU; /* Values are 12-bit */
    factor = g_avrVoltageScaling[avrVoltageIndex].factor;
    value = ((int16_t)val16 * factor) >> 10;
    pReading->doubleValue.intPart = (value >> 16) & 0xFFFFU;
    pReading->doubleValue.fracPart = ((uint32_t)value & 0xFFFFU) << 16;
    return CoreSensorSuccess;
  }
  return CoreSensorInvalidIndex;
}

static CoreSensorStatus
readSensorAvrTemp(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int avrTempIndex,
  CoreSensorValue* pReading)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint16_t val16;
  int32_t factor, offset, value;
  unsigned int fieldOffset, fieldMask;
  boolean_t bFieldSigned;

  dfAssert(DF_ARRAY_LENGTH(g_avrTempInfo) == DF_ARRAY_LENGTH(g_avrTempScaling));

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  if (avrTempIndex < DF_ARRAY_LENGTH(g_avrTempInfo)) {
    fieldOffset = g_avrTempScaling[avrTempIndex].fieldOffset;
    fieldMask = g_avrTempScaling[avrTempIndex].fieldMask;
    bFieldSigned = g_avrTempScaling[avrTempIndex].bFieldSigned;
    factor = g_avrTempScaling[avrTempIndex].factor;
    offset = g_avrTempScaling[avrTempIndex].offset;
    val16  = (uint16_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 0) << 0;
    val16 |= (uint16_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 1) << 8;
    val16 &= fieldMask;
    if (bFieldSigned) {
      /* Raw ADC value is signed (two's complement); assumes that bit 15 of val16 is sign bit */
      value = ((int32_t)(int16_t)val16 * factor) >> 4;
    } else {
      /* Raw ADC value is unsigned */
      value = ((uint32_t)val16 * factor) >> 4;
    }
    value += offset;
    pReading->doubleValue.intPart = (value >> 16) & 0xFFFFU;
    pReading->doubleValue.fracPart = ((uint32_t)value & 0xFFFFU) << 16;
    return CoreSensorSuccess;
  }
  return CoreSensorInvalidIndex;
}

static CoreSensorStatus
readSensorAvrMisc(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int avrMiscIndex,
  CoreSensorValue* pReading)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t val32;
  unsigned int fieldOffset;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  switch (avrMiscIndex) {
  case 0:
    fieldOffset = 0;
    val32  = (uint32_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 0) << 0;
    val32 |= (uint32_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 1) << 8;
    val32 |= (uint32_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 2) << 16;
    val32 |= (uint32_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 3) << 24;
    pReading->doubleValue.intPart = (int32_t)(val32 >> 2);
    pReading->doubleValue.fracPart = (val32 & 0x3) << 30;
    break;

  case 1:
    fieldOffset = 6;
    val32  = (uint32_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 0) << 0;
    val32 |= (uint32_t)dfPciMemRead8(hModel, pModelRegs->model3.avr.sensor + fieldOffset + 1) << 8;
    pReading->uint32Value = val32;
    break;

  default:
    return CoreSensorInvalidIndex;
  }

  return CoreSensorSuccess;
}

static unsigned int
initSensorInfoAvrVolt(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int sensorIndex)
{
  const unsigned int m = pDevCtx->model.admxrc6tadv8.avr.numVoltageSensor;
  unsigned int i, n = 0;

  dfAssert(sensorIndex + m <= MAX_NUM_SENSOR);
  for (i = 0; i < m && sensorIndex < MAX_NUM_SENSOR; i++, sensorIndex++) {
    dfStrCpy(pDevCtx->info.sensor[sensorIndex].description, DF_ARRAY_LENGTH(pDevCtx->info.sensor[sensorIndex].description), g_avrVoltageInfo[i].pDescription);
    pDevCtx->info.sensor[sensorIndex].capabilities = g_avrVoltageInfo[i].capabilities;
    pDevCtx->info.sensor[sensorIndex].error.intPart = g_avrVoltageInfo[i].error.intPart;
    pDevCtx->info.sensor[sensorIndex].error.fracPart = g_avrVoltageInfo[i].error.fracPart;
    pDevCtx->info.sensor[sensorIndex].datatype = g_avrVoltageInfo[i].datatype;
    pDevCtx->info.sensor[sensorIndex].unit = g_avrVoltageInfo[i].unit;
    pDevCtx->info.sensor[sensorIndex].exponent = g_avrVoltageInfo[i].exponent;
    n++;
  }
  return n; 
}

static unsigned int
initSensorInfoAvrTemp(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int sensorIndex)
{
  unsigned int i, n = 0;

  dfAssert(sensorIndex + DF_ARRAY_LENGTH(g_avrTempInfo) <= MAX_NUM_SENSOR);
  for (i = 0; i < DF_ARRAY_LENGTH(g_avrTempInfo) && sensorIndex < MAX_NUM_SENSOR; i++, sensorIndex++) {
    dfStrCpy(pDevCtx->info.sensor[sensorIndex].description, DF_ARRAY_LENGTH(pDevCtx->info.sensor[sensorIndex].description), g_avrTempInfo[i].pDescription);
    pDevCtx->info.sensor[sensorIndex].capabilities = g_avrTempInfo[i].capabilities;
    pDevCtx->info.sensor[sensorIndex].error.intPart = g_avrTempInfo[i].error.intPart;
    pDevCtx->info.sensor[sensorIndex].error.fracPart = g_avrTempInfo[i].error.fracPart;
    pDevCtx->info.sensor[sensorIndex].datatype = g_avrTempInfo[i].datatype;
    pDevCtx->info.sensor[sensorIndex].unit = g_avrTempInfo[i].unit;
    pDevCtx->info.sensor[sensorIndex].exponent = g_avrTempInfo[i].exponent;
    n++;
  }
  return n; 
}

static unsigned int
initSensorInfoMisc(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int sensorIndex)
{
  unsigned int i, n = 0;

  dfAssert(sensorIndex + DF_ARRAY_LENGTH(g_avrMiscInfo) <= MAX_NUM_SENSOR);
  for (i = 0; i < DF_ARRAY_LENGTH(g_avrMiscInfo) && sensorIndex < MAX_NUM_SENSOR; i++, sensorIndex++) {
    dfStrCpy(pDevCtx->info.sensor[sensorIndex].description, DF_ARRAY_LENGTH(pDevCtx->info.sensor[sensorIndex].description), g_avrMiscInfo[i].pDescription);
    pDevCtx->info.sensor[sensorIndex].capabilities = g_avrMiscInfo[i].capabilities;
    pDevCtx->info.sensor[sensorIndex].error.intPart = g_avrMiscInfo[i].error.intPart;
    pDevCtx->info.sensor[sensorIndex].error.fracPart = g_avrMiscInfo[i].error.fracPart;
    pDevCtx->info.sensor[sensorIndex].datatype = g_avrMiscInfo[i].datatype;
    pDevCtx->info.sensor[sensorIndex].unit = g_avrMiscInfo[i].unit;
    pDevCtx->info.sensor[sensorIndex].exponent = g_avrMiscInfo[i].exponent;
    n++;
  }
  return n; 
}

static void*
enableAvrRxNotEmptyInt(
  DfInterruptObject* pInterruptObject,
  void* pArgument1,
  void* pArgument2Ignored)
{
  AvrDeviceContext* pAvrCtx = (AvrDeviceContext*)pArgument1;
	Adb3CoreDeviceContext* pDevCtx = DF_CONTAINER_OF(pAvrCtx, Adb3CoreDeviceContext, model.admxrc6tadv8.avr.context);
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  pDevCtx->model.admxrc6tadv8.shadow.intCtl |= ADMXRC6TADV8_V_INTCTL_ADVRXNE;
  dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
  dfPciMemRead32(hModel, &pModelRegs->intCtl);

  return (void*)(uintptr_t)TRUE;
}

static void*
enableAvrTxEmptyInt(
  DfInterruptObject* pInterruptObject,
  void* pArgument1,
  void* pArgument2Ignored)
{
  AvrDeviceContext* pAvrCtx = (AvrDeviceContext*)pArgument1;
	Adb3CoreDeviceContext* pDevCtx = DF_CONTAINER_OF(pAvrCtx, Adb3CoreDeviceContext, model.admxrc6tadv8.avr.context);
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  pDevCtx->model.admxrc6tadv8.shadow.intCtl |= ADMXRC6TADV8_V_INTCTL_ADVTXE;
  dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
  dfPciMemRead32(hModel, &pModelRegs->intCtl);

  return (void*)(uintptr_t)TRUE;
}

static void*
alertBitPoll(
  DfInterruptObject* pInterruptObject,
  void* pArgument1,
  void* pArgument2)
{
	Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pArgument1;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t intStat;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  intStat = dfPciMemRead32(hModel, &pModelRegs->intStat);
  if (intStat & ADMXRC6Tx_V_INTSTAT_SYS) {
    /* Alert interrupt still active */
    return (void*)(uintptr_t)TRUE;
  } else {
    /* Alert interrupt no longer active - reenable alert interrupt */
    pDevCtx->model.admxrc6tadv8.shadow.intCtl |= ADMXRC6Tx_V_INTCTL_SYSENABLE;
    dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
    dfPciMemRead32(hModel, &pModelRegs->intCtl);
    return (void*)(uintptr_t)FALSE;
  }
}

/* Returns TRUE if alert interrupt is active for specified target FPGA */
static boolean_t
alertPoll(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int targetIndex)
{
  if (targetIndex > 0) {
    return FALSE;
  }
  return (boolean_t)(uintptr_t)dfInterruptSynchronize(&pDevCtx->interrupt.object, alertBitPoll, (void*)pDevCtx, (void*)(size_t)targetIndex);
}

static ClockProgramStatus
clockProgram(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int clockIndex,
  const CoreClockWord* pClockWord)
{
  return ClockProgramInvalidIndex;
}

static ClockWordStatus
clockWord(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int clockIndex,
  uint32_t flags,
  uint64_t frequencyReq,
  uint64_t frequencyMin,
  uint64_t frequencyMax,
  CoreClockWord* pClockWord)
{
  return ClockWordInvalidIndex;
}

static FpgaControlStatus
fpgaControl(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  CoreFpgaControlOp opCode,
  boolean_t* pbValue)
{
  dfDebugPrint(8, ("fpgaControl: index=%lu opCode=%lu pbValue=%p\n",
    (unsigned long)index, (unsigned long)opCode, pbValue));
  
  if (0 != index) {
    return FpgaControlInvalidIndex;
  }

  /* ADM-XRC-6T-ADV8 cannot be fully reconfigured */
  return FpgaControlNotSupported;
}

static FpgaSelectMapStatus
fpgaSelectMap(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  boolean_t bWrite,
  size_t length,
  void* pBuffer)
{
  dfDebugPrint(8, ("fpgaSelectMap: index=%lu bWrite=%s length=%llu(0x%08lx_%08lx) pBuffer=%p\n",
    (unsigned long)index, bWrite ? "TRUE" : "FALSE", (unsigned long long)length, dfSplitUint64((uint64_t)length), pBuffer));
  
  if (0 != index) {
    return FpgaSelectMapInvalidIndex;
  }
  return FpgaSelectMapNotSupported;
}

static boolean_t
initHardware(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bInit,
  unsigned int phase)
{
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  unsigned int i, n;
  uint32_t val32;

  dfDebugPrint(2, ("initHardware: bInit=%s phase=%u\n", bInit ? "TRUE" : "FALSE", phase));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  if (bInit) {
    switch (phase) {
    case 0:
      pDevCtx->info.model = CoreModelAdmxrc6tadv8;

      /* The set of Bridge interrupts that we are interested in */
      pDevCtx->hardware.shadow.bridge.adb3.irqEnableMaster = ADB3IRQ_DMA_ALL | ADB3IRQ_MODEL;

      /* Ensure Bridge interrupts are disabled */
      pDevCtx->hardware.shadow.bridge.adb3.irqEnable = 0;
      dfPciMemWrite32(hBridge, &pRegs->irqEnable, pDevCtx->hardware.shadow.bridge.adb3.irqEnable);
      dfPciMemRead32(hBridge, &pRegs->irqEnable);

      /* Enable model-specific interrupts that we are interested in */
      pDevCtx->model.admxrc6tadv8.shadow.intCtl = ADMXRC6Tx_V_INTCTL_FPENABLE | ADMXRC6Tx_V_INTCTL_SYSENABLE;
      dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
      dfPciMemRead32(hModel, &pModelRegs->intCtl);

      /* Initialize DMA enables */
      dfPciMemWrite32(hBridge, &pRegs->dmaEnable, ADB3DMAEN_LOCAL_MASTER | ADB3DMAEN_DMA_ALL);

      /* Figure out directions supported by DMA engine registers */
      n = 0;
      for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
        /*
        ** On the ADM-XRC-6T-ADV8, there may be up to 4 DMA channels, but we
        ** don't know how many until we get here. If a DMA channel is present,
        ** the nextAddrHigh register is writable and returns the last data
        ** written.
        ** Some DMA engines might only work in a particular direction, and we
        ** detect that by checking both 16-bit halves of the FIFO level registers.
        */
        dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].nextAddrHigh, 0xDEADBEEFU);
        val32 = dfPciMemRead32(hBridge, &pRegs->dmaEngine[i].nextAddrHigh);
        if (0xDEADBEEFU != val32) {
          /* DMA engine not present; don't look for any more DMA engines. */
          break;
        }
        n++;
        val32 = dfPciMemRead32(hBridge, &pRegs->dmaEngine[i].cleanup.fifoStatus);
        if (0U == (val32 & 0x0000ffffU)) {
          /* Does not support FPGA to Host transfers */
          pDevCtx->info.bootstrap.dmaRestrictions[i].bNoFpgaToHost = 1U;
        }
        if (0U == (val32 & 0xffff0000U)) {
          /* Does not support Host to FPGA transfers */
          pDevCtx->info.bootstrap.dmaRestrictions[i].bNoHostToFpga = 1U;
        }
      }
      pDevCtx->info.bootstrap.numDmaChannel = n;

      /* Initialize DMA engine registers */
      for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
        /* Request up to 32 x 16-byte words on PCIe side */
        dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].hostControl, 0x20U);    
        /* Allow up to 64 bytes read and 64 bytes written on MPTL */
        dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].fpgaControl, 0x80008U);
      }

      /* Identify Flash device */
      if (pDevCtx->info.bootstrap.flashWidth != 0) {
        admxrc6txFlashReset(pDevCtx);
        if (flashCfiIdentify(pDevCtx, 0, 0)) {
          pDevCtx->info.flash.bPresent = TRUE;
          pDevCtx->info.flash.targetArea.start = 0U;
          pDevCtx->info.flash.targetArea.length = pDevCtx->info.flash.totalSize - pDevCtx->info.flash.targetArea.start;
        } else {
          dfDebugPrint(0, ("*** initHardware: failed to identify Flash\n"));
        }
      }

      /* Initialize I2C stuff */
      i2cInit(&pDevCtx->model.admxrc6tadv8.i2c.context, NULL, admxrc6txI2cStart, admxrc6txI2cPoll, pDevCtx);

      /* Get the VPD */
      readVpd(pDevCtx);
      pDevCtx->model.admxrc6tadv8.avr.numVoltageSensor = DF_ARRAY_LENGTH(g_avrVoltageInfo);
      /* Correct the number of sensors for board rev. 2, if necessary */
      if (pDevCtx->info.vpd.pData->admxrc6tadv8.rev0.pcbRev >= 0x20U) {
        /* Last sensor is not present on board rev. 2 */
        pDevCtx->model.admxrc6tadv8.avr.numVoltageSensor--;
        pDevCtx->info.bootstrap.numSensor--;
      }

      pDevCtx->info.prodTest.serialNumber = pDevCtx->info.vpd.pData->admxrc6tadv8.rev0.serial;
      pDevCtx->info.prodTest.cpldRevision = 0; /* TO DO - get microcontroller revision */
      pDevCtx->info.prodTest.pcbRevision = pDevCtx->info.vpd.pData->admxrc6tadv8.rev0.pcbRev;
      pDevCtx->info.prodTest.bridgeDate = dfPciMemRead32(hModel, &pModelRegs->buildDate);
      pDevCtx->info.prodTest.bridgeTime = dfPciMemRead32(hModel, &pModelRegs->buildTime);

      /* Find base addresses and sizes of local bus windows */
      i = 0;
      pDevCtx->info.window[i].busBase = pDevCtx->rawResources.variant.pci.memBar[2].base;
      pDevCtx->info.window[i].busSize = pDevCtx->rawResources.variant.pci.memBar[2].size;
      pDevCtx->info.window[i].localBase = 0;
      pDevCtx->info.window[i].localSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
      pDevCtx->info.window[i].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[2].base;
      pDevCtx->info.window[i].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
      pDevCtx->info.window[i].pKernelBase = pDevCtx->hardware.pBridged;
      pDevCtx->info.window[i].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[2].size;
      i++;
      if (NULL != pDevCtx->hardware.pBridgedPrefetch) {
        pDevCtx->info.window[i].busBase = pDevCtx->rawResources.variant.pci.memBar[3].base;
        pDevCtx->info.window[i].busSize = pDevCtx->rawResources.variant.pci.memBar[3].size;
        pDevCtx->info.window[i].localBase = 0;
        pDevCtx->info.window[i].localSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
        pDevCtx->info.window[i].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[3].base;
        pDevCtx->info.window[i].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
        pDevCtx->info.window[i].pKernelBase = pDevCtx->hardware.pBridgedPrefetch;
        pDevCtx->info.window[i].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[3].size;
        i++;
      }
      pDevCtx->info.window[i].busBase = pDevCtx->rawResources.variant.pci.memBar[1].base;
      pDevCtx->info.window[i].busSize = pDevCtx->rawResources.variant.pci.memBar[1].size;
      pDevCtx->info.window[i].localBase = 0;
      pDevCtx->info.window[i].localSize = 0;
      pDevCtx->info.window[i].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[1].base;
      pDevCtx->info.window[i].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[1].size;
      pDevCtx->info.window[i].pKernelBase = pDevCtx->hardware.model.pGeneric;
      pDevCtx->info.window[i].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
      i++;
      pDevCtx->info.window[i].busBase = pDevCtx->rawResources.variant.pci.memBar[0].base;
      pDevCtx->info.window[i].busSize = pDevCtx->rawResources.variant.pci.memBar[0].size;
      pDevCtx->info.window[i].localBase = 0;
      pDevCtx->info.window[i].localSize = 0;
      pDevCtx->info.window[i].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[0].base;
      pDevCtx->info.window[i].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[0].size;
      pDevCtx->info.window[i].pKernelBase = pDevCtx->hardware.bridge.pGeneric;
      pDevCtx->info.window[i].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[0].size;
      i++;

      /* Initialize sensor information structures */
      i = 0;
      i += initSensorInfoAvrVolt(pDevCtx, i);
      i += initSensorInfoAvrTemp(pDevCtx, i);
      i += admxrc6txInitSensorInfoXsmon(pDevCtx, i);
      i += initSensorInfoMisc(pDevCtx, i);
      break;

    case 1:
      /* Initialise the AVR uC interface */
      if (avrInit(
        &pDevCtx->model.admxrc6tadv8.avr.context,
        &pDevCtx->interrupt.object,
        &pDevCtx->hardware.model.pAdmxrc6tx->model2.avr.curr.ctlStat,
        pDevCtx->hardware.hModel,
        ADMXRC6TADV8_AVR_FLASH_SIZE,
        ADMXRC6TADV8_AVR_FLASH_PAGE_SIZE,
        enableAvrRxNotEmptyInt,
        enableAvrTxEmptyInt))
      {
        pDevCtx->model.admxrc6tadv8.bAvrInitialized = TRUE;
      }
      break;

    default:
      dfAssert(FALSE);
      break;
    }
  } else {
    switch (phase) {
    case 0:
      /* Uninitialize I2C stuff */
      i2cUninit(&pDevCtx->model.admxrc6tadv8.i2c.context);

      /* Deinitialize DMA enables */
      dfPciMemWrite32(hBridge, &pRegs->dmaEnable, 0);
      break;

    case 1:
      /* Deinitialize the AVR interface */
      if (pDevCtx->model.admxrc6tadv8.bAvrInitialized) {
        avrUninit(&pDevCtx->model.admxrc6tadv8.avr.context);
        pDevCtx->model.admxrc6tadv8.bAvrInitialized = FALSE;
      }
      break;

    default:
      dfAssert(FALSE);
      break;
    }
  }

  return TRUE;
}

static boolean_t
isr(
  DfInterruptObject* pInterruptObj,
  void* pInterruptContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterruptContext;
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t irqStatus, irqEnable, dmaIrqStatus;
  uint32_t intStat, intCtl;
  unsigned int channel, n;

  dfDebugPrint(8, ("isr: entered\n"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  /* Sample the currently active interrupts, and mask the value with the interrupts we want */
  irqEnable = pDevCtx->hardware.shadow.bridge.adb3.irqEnable;
  irqStatus = dfPciMemRead32(hBridge, &pRegs->irqStatus) & irqEnable;

  /* Clear the interrupts we saw */
  dfPciMemWrite32(hBridge, &pRegs->irqStatus, irqStatus);

  if (irqStatus & ADB3IRQ_MODEL) {
    intCtl = pDevCtx->model.admxrc6tadv8.shadow.intCtl;
    intStat = dfPciMemRead32(hModel, &pModelRegs->intStat) & intCtl;
    dfDebugPrint(9, ("isr: model-specific interrupt, intCtl=0x%08lx intStat=0x%08lx\n", intCtl, intStat));
    /* Clear the model-specific interrupts that we saw */
    dfPciMemWrite32(hModel, &pModelRegs->intStat, intStat);
    if (intStat & ADMXRC6Tx_V_INTSTAT_FP) {
      dfDebugPrint(9, ("isr: target FPGA 0 interrupt\n"));
      dfDpcSchedule(&pDevCtx->interrupt.fpga[0].dpc, (void*)(uintptr_t)0);
    }
    if (intStat & ADMXRC6Tx_V_INTSTAT_SYS) {
      dfDebugPrint(9, ("isr: hardware monitor interrupt\n"));
      dfDpcSchedule(&pDevCtx->interrupt.hwMon[0].dpc, (void*)(uintptr_t)0);
      /* Disable hardware monitor interrupt */
      pDevCtx->model.admxrc6tadv8.shadow.intCtl &= ~ADMXRC6Tx_V_INTCTL_SYSENABLE;
      dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
      dfPciMemRead32(hModel, &pModelRegs->intCtl);
    }
    if (intStat & ADMXRC6TADV8_V_INTSTAT_ADVRXNE) {
      dfDebugPrint(9, ("isr: AVR RX FIFO not empty interrupt\n"));
      dfDpcSchedule(&pDevCtx->model.admxrc6tadv8.avr.context.transport.rx.dpc, NULL);
      /* Disable RX FIFO not empty interrupt */
      pDevCtx->model.admxrc6tadv8.shadow.intCtl &= ~ADMXRC6TADV8_V_INTSTAT_ADVRXNE;
      dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
      dfPciMemRead32(hModel, &pModelRegs->intCtl);
    }
    if (intStat & ADMXRC6TADV8_V_INTSTAT_ADVTXE) {
      dfDebugPrint(9, ("isr: AVR TX FIFO empty interrupt\n"));
      dfDpcSchedule(&pDevCtx->model.admxrc6tadv8.avr.context.transport.tx.dpc, NULL);
      /* Disable TX FIFO empty interrupt */
      pDevCtx->model.admxrc6tadv8.shadow.intCtl &= ~ADMXRC6TADV8_V_INTSTAT_ADVTXE;
      dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
      dfPciMemRead32(hModel, &pModelRegs->intCtl);
    }
  }

  n = pDevCtx->info.bootstrap.numDmaChannel;
  for (channel = 0; channel < n; channel++) {
    if (irqStatus & (ADB3IRQ_DMA0 << channel)) {
      dmaIrqStatus = dfPciMemRead32(hBridge, &pRegs->dmaEngine[channel].irq.status);
      dfDebugPrint(9, ("isr: DMA%lu interrupt, DMA%lu.irqStatus=0x%08lx\n",
        (unsigned long)channel, (unsigned long)channel, (unsigned long)dmaIrqStatus));
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[channel].irq.ack, dmaIrqStatus);
      dfDpcSchedule(&pDevCtx->interrupt.dma[channel].dpc, (void*)(uintptr_t)channel);
    }
  }

  return irqStatus ? TRUE : FALSE;
}

static boolean_t
mapBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated,
	boolean_t bUnmap)
{
	dfDebugPrint(1, ("mapBusResources: bUnmap=%s\n", bUnmap ? "TRUE" : "FALSE"));

	if (bUnmap) {
		if (NULL != pDevCtx->hardware.bridge.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.bridge.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 generic registers\n"));
			}
      pDevCtx->hardware.bridge.pGeneric = NULL;
		}
		if (NULL != pDevCtx->hardware.model.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.model.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 model-specific registers\n"));
			}
      pDevCtx->hardware.model.pGeneric = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridged) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridged)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 bridged space\n"));
			}
      pDevCtx->hardware.pBridged = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridgedPrefetch) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridgedPrefetch)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 prefetchable bridged space\n"));
			}
      pDevCtx->hardware.pBridgedPrefetch = NULL;
      pDevCtx->info.bootstrap.numWindow--;
		}
		dfDebugPrint(2, ("mapBusResources: unmapped\n"));
	} else {
		pDevCtx->hardware.bridge.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 0, 0, pTranslated->variant.pci.memBar[0].size, &pDevCtx->hardware.hBridge);
		if (NULL == pDevCtx->hardware.bridge.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 generic registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 generic registers @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[0].base), pDevCtx->hardware.bridge.pGeneric));
		}
		pDevCtx->hardware.model.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 1, 0, pTranslated->variant.pci.memBar[1].size, &pDevCtx->hardware.hModel);
		if (NULL == pDevCtx->hardware.model.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 model-specific registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 model-specific registers @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[1].base), pDevCtx->hardware.model.pGeneric));
		}
    pDevCtx->hardware.pBridged = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 2, 0, pTranslated->variant.pci.memBar[2].size, &pDevCtx->hardware.hBridged);
		if (NULL == pDevCtx->hardware.pBridged) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 bridged space @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[2].base), pDevCtx->hardware.pBridged));
		}
    if (pTranslated->variant.pci.numMemBar >= 4) {
      pDevCtx->hardware.pBridgedPrefetch = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 3, 0, pTranslated->variant.pci.memBar[3].size, &pDevCtx->hardware.hBridgedPrefetch);
      if (NULL == pDevCtx->hardware.pBridgedPrefetch) {
        dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 prefetchable bridged space\n"));
        return FALSE;
      } else {
        dfDebugPrint(2, ("mapBusResources: mapped ADB3 prefetchable bridged space @ 0x%08lx_%08lx to %p\n",
          dfSplitUint64(pTranslated->variant.pci.memBar[3].base), pDevCtx->hardware.pBridgedPrefetch));
        pDevCtx->info.bootstrap.numWindow++;
      }
    }
		dfDebugPrint(2, ("mapBusResources: mapped OK\n"));
	}

	return TRUE;
}

static void
powerChange(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int currentState,
  unsigned int newState)
{
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  unsigned int i;

  dfAssert(newState == 0 || newState == 3);

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  dfDebugPrint(1, ("powerChange: D%u->D%u\n", currentState, newState));
  if (currentState != newState) {
    if (currentState < newState) { /* Powering down (to D0) */
      /* Nothing to do yet */
    } else { /* Powering up (from D3) */
      /* Reenable model-specific interrupts that we are interested in */
      pDevCtx->model.admxrc6tadv8.shadow.intCtl = ADMXRC6Tx_V_INTCTL_FPENABLE | ADMXRC6Tx_V_INTCTL_SYSENABLE;
      dfPciMemWrite32(hModel, &pModelRegs->intCtl, pDevCtx->model.admxrc6tadv8.shadow.intCtl | ADMXRC6TADV8_V_INTCTL_LEVELBITS);
      dfPciMemRead32(hModel, &pModelRegs->intCtl);

      /* Reinitialize DMA enables */
      dfPciMemWrite32(hBridge, &pRegs->dmaEnable, ADB3DMAEN_LOCAL_MASTER | ADB3DMAEN_DMA_ALL);

      if (pDevCtx->pDevObj->hardwareId.variant.pci.revision < 0x04U) {
        /* Reinitialize PCI-E packet control for rev 0x03 or lower; allow up to 3 PCIe packets in flight at a time */
        dfPciMemWrite32(hBridge, &pRegs->pcieControl, 0x02100100U);           
      }

      /* Reinitialize DMA engine registers */
      for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
        /* Request up to 32 x 16-byte words on PCIe side */
        dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].hostControl, 0x20U);    
        /* Allow up to 64 bytes read and 64 bytes written on MPTL */
        dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].fpgaControl, 0x80008U);
      }

      /* Reset Flash device */
      admxrc6txFlashReset(pDevCtx);
    }
  }
}

static CoreSensorStatus
readSensor(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  CoreSensorValue* pReading)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  CoreSensorStatus status = CoreSensorInvalidIndex;

  if (index >= pDevCtx->info.bootstrap.numSensor) {
    return CoreSensorInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  status = readSensorAvrVoltage(pDevCtx, index, pReading);
  if (status == CoreSensorSuccess) {
    return status;
  } else if (status != CoreSensorInvalidIndex) {
    return status;
  } else {
    index -= pDevCtx->model.admxrc6tadv8.avr.numVoltageSensor;
  }

  status = readSensorAvrTemp(pDevCtx, index, pReading);
  if (status == CoreSensorSuccess) {
    return status;
  } else if (status != CoreSensorInvalidIndex) {
    return status;
  } else {
    index -= DF_ARRAY_LENGTH(g_avrTempInfo);
  }

  status = admxrc6txReadSensorXsmon(pDevCtx, index, pReading);
  if (status == CoreSensorSuccess) {
    return status;
  } else if (status != CoreSensorInvalidIndex) {
    return status;
  } else {
    index -= ADMXRC6TX_NUM_XSMON_SENSOR;
  }

  status = readSensorAvrMisc(pDevCtx, index, pReading);
  if (status == CoreSensorSuccess) {
    return status;
  } else if (status != CoreSensorInvalidIndex) {
    return status;
  } else {
    index -= DF_ARRAY_LENGTH(g_avrMiscInfo);
  }

  return status;
}

static CoreVpdStatus
mapAvrFlashStatus(
  AvrFlashStatus status)
{
  switch (status) {
  case AvrFlashSuccess:
    return CoreVpdSuccess;

  case AvrFlashTimeout:
    return CoreVpdTimeout;

  case AvrFlashNotServiceMode:
    return CoreVpdNotPresent;

  case AvrFlashInvalidRegion:
    return CoreVpdInvalidRegion;

  case AvrFlashGeneralFailure:
    return CoreVpdGeneralFailure;

  default:
    if (status >= AvrFlashHardwareFailure && status < AvrFlashHardwareFailure + 0x100U) {
      return CoreVpdGeneralFailure;
    }
    return CoreVpdGeneralFailure;
  }
}

static CoreVpdStatus
readWriteVpd(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bWrite,
  size_t address,
  size_t length,
  void* pData)
{
  AvrFlashStatus status;

  if (CHECK_BOUNDS_OFFSET(address, length, pDevCtx->info.bootstrap.vpdOffset, pDevCtx->info.bootstrap.vpdLength)) {
    return CoreVpdInvalidRegion;
  }
  if (bWrite) {
    /* Casts are safe because of above bounds check */
    status = avrFlashWrite(pDevCtx, &pDevCtx->model.admxrc6tadv8.avr.context, (uint32_t)address, (uint32_t)length, pData);
  } else {
    /* Casts are safe because of above bounds check */
    status = avrFlashRead(pDevCtx, &pDevCtx->model.admxrc6tadv8.avr.context, (uint32_t)address, (uint32_t)length, pData);
  }
  return mapAvrFlashStatus(status);
}

static boolean_t
validateBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated)
{
	dfDebugPrint(1, ("validateBusResources: entered\n"));

	if (pTranslated->variant.pci.numMemBar < 3) {
		dfDebugPrint(0, ("*** validateBusResources: less than 4 memory BARs\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[0].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR0 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[1].size < sizeof(ModelRegsAdmxrc6tx)) {
		dfDebugPrint(0, ("*** validateBusResources: BAR1 smaller than 0x%lu bytes\n", (unsigned long)sizeof(ModelRegsAdmxrc6tx)));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[2].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
  if (pTranslated->variant.pci.numMemBar >= 4 && pTranslated->variant.pci.memBar[3].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR3 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.numIrq != 1) {
		dfDebugPrint(0, ("*** validateBusResources: not exactly 1 interrupt\n"));
		return FALSE;
	}

	dfDebugPrint(2, ("validateBusResources: OK\n"));

	return TRUE;
}

void
adb3CoreBootstrapAdmxrc6tadv8(
	Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreBootstrapAdmxrc6tadv8: entered\n"));

  dfAssert(DF_OFFSET_OF(VpdAdmxrc6tadv8Rev0, checksum) == sizeof(VpdAdmxrc6tadv8Rev0) - 1);

  pDevCtx->methods.pAlertPoll = alertPoll;
  pDevCtx->methods.pClockProgram = clockProgram;
  pDevCtx->methods.pClockWord = clockWord;
  pDevCtx->methods.pDmaList = adb3DmaList;
  pDevCtx->methods.pDmaTransfer = adb3DmaTransfer;
  pDevCtx->methods.pEnableInterrupts = adb3EnableInterrupts;
  pDevCtx->methods.pFlash = admxrc6txFlash;
  pDevCtx->methods.pFpgaControl = fpgaControl;
  pDevCtx->methods.pFpgaSelectMap = fpgaSelectMap;
  pDevCtx->methods.pInitHardware = initHardware;
  pDevCtx->methods.pIsr = isr;
  pDevCtx->methods.pMapBusResources = mapBusResources;
  pDevCtx->methods.pPowerChange = powerChange;
  pDevCtx->methods.pReadSensor = readSensor;
  pDevCtx->methods.pReadWriteVpd = readWriteVpd;
  pDevCtx->methods.pSyncVpd = NULL;
  pDevCtx->methods.pValidateBusResources = validateBusResources;

  pDevCtx->info.bootstrap.numClockGen = 0;
  pDevCtx->info.bootstrap.numDmaChannel = ADMXRC6TADV8_NUM_DMA_CHANNEL;
  pDevCtx->info.bootstrap.numIoModule = 0;
  pDevCtx->info.bootstrap.numSensor = ADMXRC6TX_NUM_XSMON_SENSOR + DF_ARRAY_LENGTH(g_avrVoltageInfo) + DF_ARRAY_LENGTH(g_avrTempInfo) + DF_ARRAY_LENGTH(g_avrMiscInfo);
  pDevCtx->info.bootstrap.numTargetFpga = 1;
  pDevCtx->info.bootstrap.numWindow = 3; /* May be incremented in initHardware method */
  pDevCtx->info.bootstrap.bVpdInFlash = FALSE;
  pDevCtx->info.bootstrap.vpdLength = ADMXRC6TADV8_AVR_FLASH_SIZE;
  pDevCtx->info.bootstrap.vpdOffset = 0;
  pDevCtx->info.bootstrap.vpdBufferLength = sizeof(VpdAdmxrc6tadv8Rev0);
  pDevCtx->info.bootstrap.flashDeviceWidth = 2;
  pDevCtx->info.bootstrap.flashWidth = 2;
  pDevCtx->info.bootstrap.b64BitDmaAddress = FALSE;
  dfParameterGetBoolean(pDevCtx->pDevObj->pDrvObj, "PciAddress64Bit", &pDevCtx->info.bootstrap.b64BitDmaAddress);
  pDevCtx->info.bootstrap.dmaDescriptorSize = 32;
  pDevCtx->info.bootstrap.dmaTransferMaxSize = 0x1000000U;
  pDevCtx->info.bootstrap.bWorkAround4kCrossing = (pDevCtx->pDevObj->hardwareId.variant.pci.revision <= 0x01U) ? TRUE : FALSE;
  pDevCtx->info.bootstrap.bI2cInterruptDriven = FALSE;
}
